#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=E1101
"""Saltstack REST API handler"""

from __future__ import print_function

import sys
import json
import logging
import time
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter

from requests.packages.urllib3.exceptions import InsecureRequestWarning
from requests.packages.urllib3.util.retry import Retry


class SaltReturnParseError(Exception):
    """Raised when saltypie is unable to parse the returned content of an API call."""
    pass


class Salt(object):
    """
    Salt's Rest API handler

    Args:
        url (str): Salt API web server URL
        username (str): The username to be used to authenticate to the salt-api
        passwd (str): The password to be used to authenticate to the salt-api
        eauth (str): External authentication method.
        trust_host (bool): Whether or not to verify host certificates.
    """

    CLIENT_LOCAL = 'local'
    CLIENT_RUNNER = 'runner'
    CLIENT_WHEEL = 'wheel'

    OUTPUT_RAW = 'raw'
    OUTPUT_JSON = 'json'
    OUTPUT_DICT = 'dict'

    def __init__(self, url, username=None, passwd=None, eauth='pam', trust_host=False):
        self.url = url
        self.username = username
        self.password = passwd
        self.eauth = eauth
        self.trust_host = trust_host
        self.token = None
        self.timeout = 60
        self.session = self._new_session()
        self.lookup_interval = 1
        self.max_retries_if_aborted = 3

        self.log = logging.getLogger(__name__)

    def _new_session(self):
        """
        Creates a new requests Session

        Returns:
            request.Session
        """
        session = requests.Session()
        session.verify = not self.trust_host
        retries = Retry(total=3, backoff_factor=5)
        session.mount(self.url, HTTPAdapter(max_retries=retries))
        return session

    def get(self, path, headers=None, timeout=None):
        """
        Executes URL calls using the GET method

        Args:
            path (str): Path to append to the URL
            headers (dict): Extra headers to be passed to the GET call
            timeout (int): The amount of seconds to wait before the request times out

        Returns:
            requests.models.Response
        """
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        try:
            ret = self.session.get(
                url=urljoin(self.url, path),
                headers=headers,
                timeout=timeout or self.timeout
            )
            return ret
        except Exception as exc:
            self.log.error(exc)
            exit(1)

    def post(self, data, path='', headers=None, timeout=None):
        """
        Executes URL calls using the POST method

        Args:
            data (dict): The data to be passed to the POST call
            path (str): The path to be appended to the URL
            headers (dict): Extra headers to be passed to the POST call
            timeout (int): The amount of seconds to wait before the request times out

        Returns:
            requests.models.Response
        """
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        retries = self.max_retries_if_aborted or 1

        while True:
            try:
                ret = self.session.post(
                    url=urljoin(self.url, path),
                    json=data,
                    headers=headers,
                    timeout=timeout or self.timeout
                )
                return ret
            except Exception:
                exc_type, exc_value, _ = sys.exc_info()

                if 'RemoteDisconnected' in str(exc_value) and retries:
                    retries -= 1
                    self.log.warning(
                        'Execution failed because of connection abortion. Remaining retries: %s out of %s...',
                        retries,
                        self.max_retries_if_aborted
                    )
                    time.sleep(5)
                else:
                    self.log.error('Exception type: %s. Exception message: %s', exc_type, exc_value)
                    exit(1)

    def login(self, eauth=None):
        """
        Creates an authenticated session with a salt API server

        Args:
            eauth (str): External authentication method.

        Returns:

        """
        eauth = eauth or self.eauth
        self.log.debug('Authenticating to salt-api using `%s` external authentication...', eauth)
        data = {
            'username': self.username,
            'password': self.password,
            'eauth': eauth
        }

        try:
            ret = self.post(path='login', data=data)
            self.token = ret.json()['return'][0]['token']
            self.session.headers.update({'X-Auth-Token': self.token})
            self.log.debug('Authentication succeed.')
            return ret.content
        except Exception:
            self.log.error('Unable to connect to salt-api at `%s`. Return code: %s', ret.url, ret.status_code)
            if ret.status_code == 503:
                self.log.error('Ensure that the salt-master services is running.')
            raise

    def execute(self, fun, client=None, target=None, tgt_type=None, args=None, kwargs=None, pillar=None,
                run_async=False, async_wait=False, output='dict', returner=None):
        """
        Executes a function using the salt API.

        Args:
            fun (str): Name of the function to be executed
            client (str): Type of client to use (local, runner or wheel).
                If not specified, the `local` client will be used.
            target (str): The minions that should execute this function.
                If the client is not set to `local`, this parameter will be ignored.
            tgt_type (str): Targeting type
            args (list): List of arguments to be passed to the function.
            kwargs (dict): Dictionary of arguments to be passed to the function.
            pillar (dict): Dictionary of pillar values to be passed to the function. Example: pillar={'sleep': 30}
            run_async (bool): Whether or not to execute the function asynchronously.
                If set to `True`, it will return the job ID.
                If the client is set to `wheel`, this parameter will be ignored.
            async_wait (bool): If this parameter is set to `True`, it will use async clients, just like the `run_async`
                parameter would, but instead of returning the job ID, it will use it to keep pulling the job result
                until it is completed.
                It is basically the same as passing `run_async=True` and calling `Salt.lookup_job`
                with `until_complete=True`.
            output (str): The output format of this method (See output constants specified in this class).
            returner (str): Which salt returner to be passed to the function.

        Returns:
            str, dict: Depends on the format passed to the `output` parameter.
        """

        if not self.token:
            self.login()

        client = client or Salt.CLIENT_LOCAL
        client_type = client

        if run_async or async_wait and (client != Salt.CLIENT_WHEEL):
            client += '_async'

        data = {
            'client': client,
            'fun': fun,
        }

        if target:
            data.update({'tgt': target})
        if tgt_type:
            data.update({'tgt_type': tgt_type})
        if pillar:
            args = args or []
            args.append('pillar={}'.format(json.dumps(pillar)))
        if args:
            data.update({'arg': args})
        if kwargs:
            if client_type == Salt.CLIENT_LOCAL:
                data.update({'kwarg': kwargs})
            else:
                data.update(**kwargs)

        if returner:
            data.update({'ret': returner})

        self.log.debug('Executing salt command: %s', data)
        ret = self.post(data)
        try:
            content_dict = json.loads(ret.content)
        except json.decoder.JSONDecodeError as exc:
            msg = 'Error: Unable to parse API return as JSON: {}. Returned code:{}. Returned content: {}'.format(
                exc, ret.status_code, ret.content)
            self.log.error(msg)
            raise SaltReturnParseError(msg)

        if async_wait:
            try:
                jid = content_dict['return'][0]['jid']
                return self.lookup_job(jid, until_complete=True, output=output)
            except KeyError as exc:
                self.log.debug(exc)
                self.log.debug('Unable to retrieve JID. Assuming no jobs were executed.')
                if target:
                    self.log.debug('Targeting might have matched no minion.')
                return dict()

        # if run_async:
        #     return json.loads(ret.content)

        # if output == Salt.OUTPUT_DICT:
        #     return json.loads(ret.content)

        if output == Salt.OUTPUT_RAW:
            return ret.content

        if output == Salt.OUTPUT_JSON:
            return ret.json()

        return content_dict


    def wheel(self, *args, **kwargs):
        """
        Used to send wheel commands to the salt master.

        See:
            Salt.execute method.
        """

        return self.execute(client=Salt.CLIENT_WHEEL, *args, **kwargs)

    def runner(self, *args, **kwargs):
        """
        Used to send commands to be executed by the salt master.

        See:
            Salt.execute method.
        """

        return self.execute(client=Salt.CLIENT_RUNNER, *args, **kwargs)

    def runner_async(self, wait=False, *args, **kwargs):
        """
        Used to send commands to be executed by the salt master asynchronously.

        Args:
            wait (bool, optional): Defaults to False. Whether or not to pool for the job until completed.

        See:
            Salt.execute method.
        """

        return self.runner(run_async=True, async_wait=wait, *args, **kwargs)

    def local(self, *args, **kwargs):
        """
        Used to send commands to be exected by salt minions.

        See:
            Salt.execute method.
        """

        return self.execute(client=Salt.CLIENT_LOCAL, *args, **kwargs)

    def local_async(self, wait=False, *args, **kwargs):
        """
        Used to send commands to be executed by salt minions asynchronously.

        Args:
            wait (bool, optional): Defaults to False. Whether or not to pool for the job until completed.

        See:
            Salt.execute method.
        """

        return self.local(run_async=True, async_wait=wait, *args, **kwargs)

    def lookup_job(self, jid, until_complete=False, interval=None, output='dict'):
        """
        Retrieves information about a saltstack job.

        Args:
            jid (int): The job ID.
            until_complete (bool): Whether or not to keep pulling the job until it is completed.
            interval (int): Time interval (in seconds) to wait in between checks.
                If not specified, defaults will be used.
                You can also set the global lookup interval using `self.lookup_interval`.
            output (str): The output format of this method (See output constants specified in this class).
        Returns:
            dict: a dictionary containing the job results
        """
        if not self.token:
            self.login()

        ret = self.execute(client=Salt.CLIENT_RUNNER, fun='jobs.lookup_jid', kwargs={'jid': jid}, output=output)

        if until_complete:
            while ret['return'][0] == {}:
                time.sleep(interval or self.lookup_interval)
                ret = self.lookup_job(jid)

        return ret

    def highstate(self, target, output='dict'):
        """
        Runs highstate on the target minion

        Args:
            target (str): The minion to run the highstate
            output (str): The output format of this method (See output constants specified in this class).
        Returns:
            str, dict: Depends on the format passed to the `output` parameter.
        """
        return self.execute(fun='state.apply', target=target, async_wait=True, output=output)

    def versions(self):
        """
        Returns the versions of salt running on master and its minions

        Returns:
            dict
        """
        ret = self.execute(fun='manage.versions', client=Salt.CLIENT_RUNNER)['return'][0]
        ver = {
            'master': ret['Master'],
            'minions': {}
        }

        if 'Up to date' in ret.keys():
            ver['minions'].update(ret['Up to date'])

        if 'Minion requires update' in ret.keys():
            ver['minions'].update(ret['Minion requires update'])
        return ver



# -*- coding: utf-8 -*-
"""Base output handler"""
import sys
import logging
from terminaltables import AsciiTable, SingleTable


class BaseOutput(object):
    """Output handler for salt execution return objects"""

    def __init__(self, ret):
        """

        Args:
            ret (dict): The dictionary the is returned as a result of running salt command via REST API.
        """
        self.log = logging.getLogger('{}.{}'.format(self.__module__, type(self).__name__))
        self.raw_data = ret
        self.data = self.ordered_result(ret)
        self.total_ms = {}
        self.colored = True

        if sys.stdout.encoding == 'utf-8':
            self.safe = False
        else:
            self.safe = True

    @staticmethod
    def extract_id(key):
        """
        Extracts a descriptive portion of an execution ID naming pattern.

        Args:
            key (str): The execution ID from which to extract the description

        Returns:
            str: The description of the execution
        """

        try:
            _id = key.split('_|-')[1]
        except IndexError:
            _id = key
        return _id

    def ordered_result(self, result):
        """Returns an ordered dictionary of the execution result

        Args:
            result (dict): The execution result dictionary to be ordered

        Returns:
            OrderedDict
        """
        pass

    @staticmethod
    def format_duration(duration):
        """Formats duration into a more human readable value.

        Args:
            duration (float): Duration time in milliseconds.

        Returns:
            str: The formatted duration.
        """

        seconds = duration / 1000
        if seconds > 60:
            formatted_duration = '{0:1.2f}min'.format(seconds / 60)
        elif seconds >= 1:
            formatted_duration = '{0:1.2f}s'.format(seconds)
        else:
            formatted_duration = '{0:1.2f}ms'.format(duration)

        return formatted_duration

    def _plot_duration(self, duration, total_duration, max_bar_size=30):
        """Returns a bar and the percentage value of the duration in relation to the total.

        Args:
            duration (float): The duration.
            total_duration (float): The total duration (Usually of a highstate).
            max_bar_size (int, optional): Defaults to 30. The bar size that corresponds to 100%.

        Returns:
            tuple: the bar, the percentage value
        """

        ticks = ('█', '▌', '|')
        plot_bar = ''
        percentage = 0

        if self.safe:
            the_tick = ticks[2]
        else:
            the_tick = ticks[0]

        try:
            factor = max_bar_size * duration / total_duration
            plot_bar = str(the_tick * int(factor))
            percentage = factor * 100 / max_bar_size
        except ZeroDivisionError:
            self.log.warning('Unable to format zero duration.')

        return plot_bar, '{0:>5.2f}%'.format(percentage)

    def format_time(self, value, unit):
        """Converts milliseconds to the specified unit.

        Args:
            value (float): The time in milliseconds.
            unit (str): The unit to convert to. `s`: for seconds, `min`: for minutes.

        Returns:
            str: Formatted time
        """

        formatted_value = value

        if unit == 's':
            formatted_value = value / 1000
        elif unit == 'min':
            formatted_value = value / 3600000
        elif unit == 'ms':
            return value
        return '{0:>5.2f}'.format(formatted_value)

    def _create_table(self, data, title=None):
        """Creates a console printable table based on the provided data.

        Args:
            data (list): List of data (As expected by terminaltables's table classes).
            title (str, optional): The table title.

        Returns:
            str: A console printable table.
        """

        if self.safe:
            table = AsciiTable(data)
        else:
            table = SingleTable(data)

        if title:
            table.title = ' {} '.format(title)

        table.inner_column_border = False
        table.inner_footing_row_border = True

        return table.table

    def set_color(self, color_method, items):
        """Sets a terminal color for all items in a list

        Args:
            color_method: A color method from the Color class to be used to format the list items.
            items (list): A list of items to be colored.
        Returns:
            list
        """
        if self.colored:
            return [color_method(item, auto=True) for item in items]
        return items



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Salt orchestration output parser"""

from collections import OrderedDict
from colorclass import Color

from saltypie.output import BaseOutput, StateOutput


class OrchestrationOutput(BaseOutput):
    """Output handler for salt orchestration return objects

    Args:
        ret (dict): The return object of a `state.orch` execution.
        salt (Salt): saltypie's Salt object for connecting to the master where the orchestration
            was executed from. A bug on salt's orcherstration return object prevents it from been
            parsed as a full qualified JSON object when one of its state execution steps fails
    """

    def __init__(self, ret, salt=None):
        super(OrchestrationOutput, self).__init__(ret)
        self.salt = salt
        self.parsed_data = self.parse_data()

    def ordered_result(self, result):
        """Order orchestration steps by run number.

        Args:
            result (dict): The return object of a `state.orch` execution.

        Returns:
            OrderedDict
        """
        ordered = {}

        self.log.debug('Ordering orchestration output...')

        try:
            data = result['return'][0]['data']
        except KeyError:
            self.log.debug('Unable to retrieve orchestration data.'
                           ' Attemping job lookup return object format...')
            data = list(result['return'][0].values())[0]['return']['data']
            self.log.debug('Data retrieved.')

        try:
            for master, _orch in data.items():
                ordered[master] = OrderedDict(
                    sorted(_orch.items(), key=lambda k: k[1]['__run_num__']))
        except Exception as exc:
            self.log.error('Error: Unable to sort orchestration results')
            self.log.error('%s: %s', type(exc), exc)
            self.log.debug('Orchestration results: \n%s', result)
            raise

        return ordered

    def normalize_state(self, state_data):
        """Normalizes an orchestration state return data by making the `changes` attribute
        consistent even and it fails.

        When normalizing data for failed states, the return object will be retrieved by querying the server
        using the job ID.

        If `OrchestrationOutput.salt` object is not provided, the state data will not be altered.

        Args:
            state_data (dict): State execution dictionary.

        Returns:
            dict: Normalized data
        """

        state_data = dict(state_data)

        if state_data['changes']:
            state_data['changes'] = {
                'return': [state_data['changes']['ret']]
            }
        else:
            if self.salt:
                state_return = self.salt.lookup_job(state_data['__jid__'])
                state_data['changes'] = state_return
            else:
                self.log.error(
                    'Error: Unable to fetch data from failed state: `%s`. Salt object not provided.',
                    self.extract_id(state_data.get('__id__'))
                )

        return state_data

    def parse_data(self, dict_only=False):
        """Parses the orchestration data.

        Args:
            dict_only (bool, optional): Defaults to False. State `changes` will be replaced by saltypie.StateOutput
                objects. If this argument is set to true, it will be just an `OrderedDict` instead.

        Returns:
            dict
        """

        ret = dict()
        for master, _orch in self.data.items():
            ret[master] = {
                'data': [],
                'total_duration': 0,
                'failed_steps': []
            }

            for key, data in _orch.items():
                ret[master]['total_duration'] += data.get('duration', 0)

                if data.get('result') is False:
                    ret[master]['failed_steps'].append(key)

                if self.is_salt_state(key):
                    try:
                        data = self.normalize_state(data)
                        state_output = StateOutput(data['changes'])
                        if dict_only:
                            data['changes'] = state_output.data
                        else:
                            data['changes'] = state_output
                    except KeyError:
                        self.log.debug('Unable to normalize data. Leaving as is.')

                ret[master]['data'].append({key: data})
        return ret

    @property
    def failed_steps(self):
        """List of all the failed steps in the orchestration.

        Returns:
            list
        """

        steps = []
        for data in self.parsed_data.values():
            steps.extend(data['failed_steps'])
        return steps

    def get_step_names(self):
        """
        List orchestration step names.

        Returns:
            list
        """

        steps = list()
        for _, _orch in self.data.items():
            for key, _ in _orch.items():
                steps.append(self.extract_id(key))
        return steps

    @staticmethod
    def get_step_type(key):
        """Returns the orcherstration step type based on its key.

        Args:
            key ([str): An orchestration step key.

        Returns:
            str: The orchestration step type.
        """

        return key.split('_|-')[-1]

    @staticmethod
    def is_salt_function(key):
        """Checks whether or not an orchestration step is a salt function execution.

        Args:
            key (str): An orchestration step key.

        Returns:
            bool
        """
        return OrchestrationOutput.get_step_type(key) == 'function'

    @staticmethod
    def is_salt_state(key):
        """Checks whether or not an orcherstration step is a salt state execution.

        Args:
            key (str): An orchestration step key

        Returns:
            bool
        """
        return OrchestrationOutput.get_step_type(key) == 'state'

    def get_state_outputs(self):
        """Returns a list of all state output objects present on the orchestration.

        Returns:
            list: List of dictionaries with two keys: step and data.
        """

        outputs = []
        for _, orch in self.parsed_data.items():
            for step in orch['data']:
                for step_name, step_data in step.items():
                    _id = self.extract_id(step_name)
                    if isinstance(step_data['changes'], StateOutput):
                        outputs.append({'step': _id, 'data': step_data['changes']})
        return outputs

    def summary_table(self, max_bar_size=30, time_unit='s', show_minions=False):
        """Returns a table listing the orchestration steps and information about its duration and result.

        Args:
            max_bar_size (int, optional): Defaults to 30. Size of the bar plot equivalent to 100%
                of the execution time.
            time_unit (str, optional): Defaults to 's'. Step duration unit.
                ms: milliseconds, s: seconds or min: minutes.

        Returns:
            str: A console printable table representation of the orchestration.
        """

        table_data = [['Step', 'Plot', '%', 'Time({})'.format(time_unit), 'Result']]

        for _, orch in self.parsed_data.items():
            for step in orch['data']:
                for step_name, step_data in step.items():
                    step_duration = step_data.get('duration', 0)
                    plot_bar, percentage = self._plot_duration(
                        duration=step_duration,
                        total_duration=orch['total_duration'],
                        max_bar_size=max_bar_size
                    )

                    _id = self.extract_id(step_name)
                    duration = self.format_time(step_duration, unit=time_unit)

                    line = (_id, plot_bar, percentage, duration, step_data['result'])
                    if step_data['result']:
                        table_data.append(self.set_color(Color.cyan, line))
                    else:
                        table_data.append(self.set_color(Color.red, line))

                    if show_minions:
                        branch = "├─ " if not self.safe else "|-- "
                        if isinstance(step_data['changes'], StateOutput):
                            for minion in step_data['changes']:
                                for minion_id, minion_data in minion.items():
                                    line = (branch + minion_id, '', '', '', minion_data['failed_states'] == [])
                                    if not minion_data['failed_states']:
                                        table_data.append(self.set_color(Color.cyan, line))
                                    else:
                                        table_data.append(self.set_color(Color.red, line))
                        elif self.is_salt_function(step_name):
                            for minion_id in step_data['changes'].get('ret', []):
                                line = (branch + minion_id, '', '', '', '')
                                if step_data['result']:
                                    table_data.append(self.set_color(Color.cyan, line))
                                else:
                                    table_data.append(self.set_color(Color.red, line))

            table_data.append([
                'Total elapsed time: {}'.format(
                    self.format_duration(orch['total_duration'])
                )
            ])

        return self._create_table(data=table_data, title='Orchestration')

    def detailed_table(self):
        """[summary]
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Salt state output parser"""

import sys
import json
from collections import OrderedDict
import colorama
from colorclass import Color, Windows

from saltypie.output.base import BaseOutput

colorama.init()

try:
    if sys.stdout.encoding == 'utf-8':
        Windows.enable(auto_colors=True, reset_atexit=True)
except AttributeError:
    pass


class StateOutput(BaseOutput):
    """Output handler for salt state return objects"""

    def ordered_result(self, result):
        """Order states by run number

        Args:
            result (dict): The return object of a `state.apply` execution

        Returns:
            OrderedDict
        """
        ordered = {}

        self.log.debug('Ordering state runs...')

        if not 'return' in result:
            self.log.debug('`return` key not found. Assuming salt-call output...')
            result = dict({
                'return': [result]
            })

        for minions in result['return']:
            for minion_id in minions:
                states = minions[minion_id]
                try:
                    ordered[minion_id] = OrderedDict(
                        sorted(states.items(), key=lambda k: k[1]['__run_num__']))
                except Exception as exc:
                    self.log.error('Error: Unable to sort state results for %s minion', minion_id)
                    self.log.error('%s: %s', type(exc), exc)
                    self.log.debug('State results: \n%s', json.dumps(result, indent=2))
                    raise

        return ordered

    def summary(self, *args, **kwargs):
        """Alias for StateOutput.parse_data to support backwards compatibility.

        See:
            StateOutput.parse_data
        """
        self.log.warning('Warning: `StateOutput.summary()` method might be removed in the future.'
                         'Use `StateOutput.parse_data()` instead.')
        return self.parse_data(*args, **kwargs)

    def parse_data(self, max_chars=None):
        """Returns the parsed data of a state run.

        ID: the state name extracted from the state key
        Duration: If a state did not run, its duration is set to zero
        Result: `True` or `False` for whether or not the state run successfully
        Changes: `True` or `False` for whether or not the state has made changes to the minion

        Args:
            max_chars (int): Maximum number of characters to display for state ID.
                If the ID is greater then `max_chars` ellipsis(...) will be added.
        Returns:
            dict
        """

        ret = {}
        for minion_id in self.data:
            ret[minion_id] = {
                'states': [],
                'total_duration': 0,
                'failed_states': [],
                'raw_states': []
            }

            for state_key, state_data in self.data[minion_id].items():

                ret[minion_id]['total_duration'] += state_data.get('duration', 0)

                state_name = BaseOutput.extract_id(state_key)
                if max_chars:
                    max_chars = abs(max_chars)
                    if len(state_name) > max_chars:
                        state_name = state_name[:max_chars] + '...'

                filtered_data = {
                    'id': state_name,
                    'duration': state_data.get('duration', 0),
                    'result': state_data['result'],
                    'changes': state_data['changes'] != {}
                }
                ret[minion_id]['raw_states'].append(state_data)
                ret[minion_id]['states'].append(filtered_data)

                if not state_data['result']:
                    ret[minion_id]['failed_states'].append(filtered_data)

        return ret

    def tables(self, failed_only=False, max_chars=None, max_bar_size=30, time_unit='ms'):
        """Creates a list of tables representing the state run including a more graphical
        representation of the duration.

        Args:
            failed_only (bool): Whether or not the tables should only contain the failed states
            max_chars (int): Maximum number of characters to display for state ID.
                If the ID is greater then `max_chars` ellipsis(...) will be added.
            max_bar_size (int): Size of the bar plot equivalent to 100% of the execution time.

        Returns:
            list
        """

        if failed_only:
            states = 'failed_states'
        else:
            states = 'states'

        data = self.parse_data(max_chars=max_chars)
        tables = []

        for minion_id in data:

            if not data[minion_id][states]:
                self.log.debug('No state results found. Skipping `%s`', minion_id)
                continue

            table_data = [['State', 'Plot', '%', 'Time({})'.format(time_unit), 'Result']]
            for state in data[minion_id][states]:

                plot_bar, percentage = self._plot_duration(
                    duration=state['duration'],
                    total_duration=data[minion_id]['total_duration'],
                    max_bar_size=max_bar_size
                )

                line = (
                    state['id'],
                    plot_bar,
                    percentage,
                    self.format_time(state['duration'], time_unit),
                    state['result']
                )

                if state['result']:
                    table_data.append(self.set_color(Color.cyan, line))
                else:
                    table_data.append(self.set_color(Color.red, line))

            table_data.append([
                'Total elapsed time: {}'.format(
                    self.format_duration(data[minion_id]['total_duration'])
                )
            ])

            tables.append(self._create_table(data=table_data, title=minion_id))

        return tables

    def graphs(self, *args, **kwargs):
        """Alias for StateOutput.tables to support backwards compatibility.

        See:
            StateOutput.tables
        """
        self.log.warning('Warning: `StateOutput.graphs()` method might be removed in the future. '
                         'Use `StateOutput.tables()` instead.')
        return self.tables(*args, **kwargs)

    def __str__(self):
        ret = ''
        for table in self.tables():
            ret += table + '\n\n'
        return ret

    def __repr__(self):
        return json.dumps(self.data)

    def __iter__(self):
        for minion_id, data in self.parse_data().items():
            minion_data = {minion_id: data}
            yield minion_data
